@@ -4,9 +4,9 @@ require Tk;
 
 
 use vars qw($VERSION);
-$VERSION = '3.004'; # $Id: //depot/Tk8/Extensions/ImageBack/ImageBack.pm#4$
+$VERSION = '3.005'; # $Id: //depot/Tk8/Extensions/ImageBack/ImageBack.pm#5$
 
-@ISA = qw(DynaLoader);
+use base  qw(DynaLoader);
 
 bootstrap Tk::ImageBack $Tk::VERSION;
 
